//-------------------------------------------------------------------------------------------------
// <copyright file="CreateProjectReferenceDefineConstants.cs" company="Microsoft">
//    Copyright (c) Microsoft Corporation.  All rights reserved.
//    
//    The use and distribution terms for this software are covered by the
//    Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
//    which can be found in the file CPL.TXT at the root of this distribution.
//    By using this software in any fashion, you are agreeing to be bound by
//    the terms of this license.
//    
//    You must not remove this notice, or any other, from this software.
// </copyright>
// 
// <summary>
// Build task to execute the compiler of the Windows Installer Xml toolset.
// </summary>
//-------------------------------------------------------------------------------------------------

namespace Microsoft.Tools.WindowsInstallerXml.Build.Tasks
{
    using System;
    using System.Collections.Generic;
    using System.IO;
    using System.Globalization;
    using System.Text;

    using Microsoft.Build.Framework;
    using Microsoft.Build.Utilities;

    /// <summary>
    /// An MSBuild task to create a list of preprocessor defines to be passed to candle from the
    /// list of referenced projects.
    /// </summary>
    public sealed class CreateProjectReferenceDefineConstants : Task
    {
        private ITaskItem[] defineConstants;
        private ITaskItem[] projectConfigurations;
        private ITaskItem[] projectReferencePaths;

        [Output]
        public ITaskItem[] DefineConstants
        {
            get { return this.defineConstants; }
        }

        [Required]
        public ITaskItem[] ProjectReferencePaths
        {
            get { return this.projectReferencePaths; }
            set { this.projectReferencePaths = value; }
        }

        public ITaskItem[] ProjectConfigurations
        {
            get { return this.projectConfigurations; }
            set { this.projectConfigurations = value; }
        }

        public override bool Execute()
        {
            List<ITaskItem> outputItems = new List<ITaskItem>();

            foreach (ITaskItem item in this.ProjectReferencePaths)
            {
                string configuration = item.GetMetadata("Configuration");
                string fullConfiguration = item.GetMetadata("FullConfiguration");
                string platform = item.GetMetadata("Platform");
                
                string projectPath = item.GetMetadata("MSBuildSourceProjectFileFullPath");
                string projectDir = Path.GetDirectoryName(projectPath) + Path.DirectorySeparatorChar;
                string projectExt = Path.GetExtension(projectPath);
                string projectFileName = Path.GetFileName(projectPath);
                string projectName = Path.GetFileNameWithoutExtension(projectPath);

                string referenceName = item.GetMetadata("Name");
                if (String.IsNullOrEmpty(referenceName))
                {
                    referenceName = projectName;
                }

                // Replace parentheses with underscores because parentheses are not valid WiX preprocessor variable characters.
                referenceName = referenceName.Replace('(', '_').Replace(')', '_');

                string targetPath = item.GetMetadata("FullPath");
                string targetDir = Path.GetDirectoryName(targetPath) + Path.DirectorySeparatorChar;
                string targetExt = Path.GetExtension(targetPath);
                string targetFileName = Path.GetFileName(targetPath);
                string targetName = Path.GetFileNameWithoutExtension(targetPath);

                // If there is no configuration metadata on the project reference task item,
                // check for any additional configuration data provided in the optional task property.
                if (String.IsNullOrEmpty(fullConfiguration))
                {
                    fullConfiguration = this.FindProjectConfiguration(projectName);
                    if (!String.IsNullOrEmpty(fullConfiguration))
                    {
                        string[] typeAndPlatform = fullConfiguration.Split('|');
                        configuration = typeAndPlatform[0];
                        platform = (typeAndPlatform.Length > 1 ? typeAndPlatform[1] : String.Empty);
                    }
                }

                // write out the platform/configuration defines
                outputItems.Add(new TaskItem(String.Format(CultureInfo.InvariantCulture, "{0}.Configuration={1}", referenceName, configuration)));
                outputItems.Add(new TaskItem(String.Format(CultureInfo.InvariantCulture, "{0}.FullConfiguration={1}", referenceName, fullConfiguration)));
                outputItems.Add(new TaskItem(String.Format(CultureInfo.InvariantCulture, "{0}.Platform={1}", referenceName, platform)));

                // write out the ProjectX defines
                outputItems.Add(new TaskItem(String.Format(CultureInfo.InvariantCulture, "{0}.ProjectDir={1}", referenceName, projectDir)));
                outputItems.Add(new TaskItem(String.Format(CultureInfo.InvariantCulture, "{0}.ProjectExt={1}", referenceName, projectExt)));
                outputItems.Add(new TaskItem(String.Format(CultureInfo.InvariantCulture, "{0}.ProjectFileName={1}", referenceName, projectFileName)));
                outputItems.Add(new TaskItem(String.Format(CultureInfo.InvariantCulture, "{0}.ProjectName={1}", referenceName, projectName)));
                outputItems.Add(new TaskItem(String.Format(CultureInfo.InvariantCulture, "{0}.ProjectPath={1}", referenceName, projectPath)));

                // write out the TargetX defines
                outputItems.Add(new TaskItem(String.Format(CultureInfo.InvariantCulture, "{0}.TargetDir={1}", referenceName, targetDir)));
                outputItems.Add(new TaskItem(String.Format(CultureInfo.InvariantCulture, "{0}.TargetExt={1}", referenceName, targetExt)));
                outputItems.Add(new TaskItem(String.Format(CultureInfo.InvariantCulture, "{0}.TargetFileName={1}", referenceName, targetFileName)));
                outputItems.Add(new TaskItem(String.Format(CultureInfo.InvariantCulture, "{0}.TargetName={1}", referenceName, targetName)));
                outputItems.Add(new TaskItem(String.Format(CultureInfo.InvariantCulture, "{0}.TargetPath={1}", referenceName, targetPath)));
            }

            this.defineConstants = outputItems.ToArray();

            return true;
        }

        /// <summary>
        /// Look through the configuration data in the ProjectConfigurations property
        /// to find the configuration for a project, if available.
        /// </summary>
        /// <param name="projectName">Name of the project that is being searched for.</param>
        /// <returns>Full configuration spec, for example "Release|Win32".</returns>
        private string FindProjectConfiguration(string projectName)
        {
            string configuration = String.Empty;

            if (this.ProjectConfigurations != null)
            {
                foreach (ITaskItem configItem in this.ProjectConfigurations)
                {
                    string configProject = configItem.ItemSpec;
                    if (configProject.Length > projectName.Length &&
                        configProject.StartsWith(projectName) &&
                        configProject[projectName.Length] == '=')
                    {
                        configuration = configProject.Substring(projectName.Length + 1);
                        break;
                    }
                }
            }

            return configuration;
        }
    }
}
